 /*
 * FSB code from the reference implementation of FSB by
 * Daniel Augot, Matthieu Finiasz, Philippe Gaborit, Stéphane Manuel, and Nicolas Sendrier.
 * http://www-rocq.inria.fr/secret/CBCrypto/fsb_reference_implementation.zip
 *
 * with modifications by Peter Schwabe
 * Pollard iteration by Peter Schwabe
 * public domain (according to communication with Matthieu Finiasz)
 */


#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>

#include "fsb.h"
#include "Whirlpool.h"
#include "nessie.h"

#define NUMBER_OF_PARAMETERS 6
extern const int parameters[NUMBER_OF_PARAMETERS][5];
extern const unsigned char Pi[2048*1024];
 
HashReturn Reset(hashState *state, int hashbitlen) 
{
  int i;

  free(state->syndrome);
  free(state->new_syndrome);
  free(state->buffer);

  for (i=0; i<NUMBER_OF_PARAMETERS; i++) {
    if (hashbitlen == parameters[i][0]) {
      state->n = parameters[i][1];
      state->w = parameters[i][2];
      state->r = parameters[i][3];
      state->p = parameters[i][4];
      state->b = state->n/state->r;
      state->bpc = logarithm(state->n/state->w);
      state->inputsize = state->w*state->bpc-state->r;
      state->bfiv = state->r/state->w;
      state->bfm = state->inputsize/state->w;
      state->hashbitlen = hashbitlen;
      state->databitlen = 0;

      state->syndrome = (unsigned char*) calloc(LUI(state->r),sizeof(unsigned int));
      state->new_syndrome = (unsigned int*) calloc(LUI(state->r),sizeof(unsigned int));
      state->buffer = (unsigned char*) malloc(state->inputsize>>3);
      state->count = 0;
      return SUCCESS;
    }
  }
  return BAD_HASHBITLEN;
}

HashReturn MyFinal(hashState * state, BitSequence *hashval) 
{
  unsigned char* padding, *whirlOutput;
  int i;
  struct NESSIEstruct * whirlpoolState;
  DataLength databitlen = state->databitlen;
  if (state->count+65 > state->inputsize) {
    padding = (unsigned char*) calloc(state->inputsize>>3,1);
    padding[0] = 1<<7;
    Update(state,padding,state->inputsize-state->count);
    padding[0] = 0;
    for (i=0; i<8; i++) {
      padding[(state->inputsize>>3)-1-i] = (unsigned char) (databitlen>>(8*i));
    }
    Update(state,padding,state->inputsize);
  } else {
    padding = (unsigned char*) calloc(((state->inputsize-state->count)>>3)+1,1);
    padding[0] = 1<<7;
    Update(state,padding,state->inputsize-state->count-64);
    for (i=0; i<8; i++) {
      padding[7-i] = (unsigned char) (databitlen>>(8*i)) ;
    }
    Update(state,padding,64);
  }
  free(padding);

  /* The final round of FSB is finished, now we simply apply the final transform: Whirlpool */
  whirlpoolState = (struct NESSIEstruct *) malloc(sizeof(struct NESSIEstruct));
  whirlOutput = (unsigned char*) malloc(64);
  NESSIEinit(whirlpoolState);

  NESSIEadd(state->syndrome,state->r,whirlpoolState);
  NESSIEfinalize(whirlpoolState, whirlOutput);

  for(i=0; i<(state->hashbitlen>>3); i++) {
    hashval[i] = whirlOutput[i];
  }

  free(whirlOutput);
  free(whirlpoolState);
  return SUCCESS;
}

void MyClean(hashState *state)
{
  int i,j;
  for (i=0; i<state->b; i++) {
    for (j=0; j<8; j++) {
      free(state->first_line[i][j]);
    }
    free(state->first_line[i]);
  }
  free(state->first_line);
  free(state->syndrome);
  free(state->new_syndrome);
  free(state->buffer);
}

HashReturn MyHash(int hashbitlen, const BitSequence *data, DataLength databitlen, BitSequence *hashval, hashState *state) {
  int return_value;
  return_value = Reset(state, hashbitlen);
  if (return_value != SUCCESS) {
    return return_value;
  }
  return_value = Update(state, data, databitlen);
  if (return_value != SUCCESS) {
    return return_value;
  }
  return_value = MyFinal(state, hashval);
  return return_value;  
}

int main(int argc, char* argv[]) {
  int hashbitlen = 48;
  int i;
  unsigned char h1[hashbitlen>>3];
  unsigned char h2[hashbitlen>>3];
  unsigned char in1[hashbitlen>>3];
  unsigned char in2[hashbitlen>>3];
  unsigned char sav[hashbitlen>>3];

  memset(in1, 42, hashbitlen>>3);
  memset(in2, 42, hashbitlen>>3);
  
  hashState state;
  Init(&state, hashbitlen);
    
  int success = 0;
  unsigned long long ctr = 0;
  unsigned long long cyclectr = 0;

  do
  {
    MyHash(hashbitlen, (const BitSequence *) in1, hashbitlen, h1, &state);


    MyHash(hashbitlen, (const BitSequence *) in2, hashbitlen, in2, &state);
    MyHash(hashbitlen, (const BitSequence *) in2, hashbitlen, h2, &state);
    ctr++;
    
    if(!memcmp(h1, h2, hashbitlen>>3)) 
      success = 1;
    memcpy(in1, h1, hashbitlen>>3);
    memcpy(in2, h2, hashbitlen>>3);
  }
  while (!success);
  memcpy(sav, h1, hashbitlen>>3);
  do
  {
    MyHash(hashbitlen, (const BitSequence *) h1, hashbitlen, h1, &state);
    cyclectr++;
  }
  while(memcmp(h1, sav, hashbitlen>>3));

  memset(in1, 42, hashbitlen>>3);
  memset(in2, 42, hashbitlen>>3);


  for(i=0;i<cyclectr;i++)
    MyHash(hashbitlen, (const BitSequence *) in2, hashbitlen, in2, &state);
  do
  {
    MyHash(hashbitlen, (const BitSequence *) in2, hashbitlen, h2, &state);
    MyHash(hashbitlen, (const BitSequence *) in1, hashbitlen, h1, &state);
    if(!memcmp(h1,h2,hashbitlen>>3))
    {
      printf("collision:\n");
      printf("%02x %02x %02x %02x %02x %02x\n",in1[0], in1[1], in1[2], in1[3], in1[4], in1[5]);
      printf("%02x %02x %02x %02x %02x %02x\n",in2[0], in2[1], in2[2], in2[3], in2[4], in2[5]);
      break;
    }
    memcpy(in1, h1, hashbitlen>>3);
    memcpy(in2, h2, hashbitlen>>3);
  }
  while(1);

  MyClean(&state);
  return 0;
}
